#include <stdio.h>
#include <ctype.h>
#include "defs.h"
#include "externs.h"
#include "protos.h"

/* ----
 * class1()
 * ----
 * 1 byte, no operand field
 */

void
class1(int *ip)
{
	check_eol(ip);

	/* update location counter */
	loccnt++;

	/* generate code */
	if (pass == LAST_PASS) {
		/* opcode */
		putbyte(data_loccnt, opval);

		/* output line */
		println();
	}
}

/* ----
 * class2()
 * ----
 * 2 bytes, relative addressing
 */

void
class2(int *ip)
{
	int addr;

	/* update location counter */
	loccnt += 2;

	/* get destination address */
	if (!evaluate(ip, ';'))
		return;

	/* generate code */
	if (pass == LAST_PASS) {
		/* opcode */
		putbyte(data_loccnt, opval);

		/* calculate branch offset */
		addr = value - (loccnt + (page << 13));

		/* check range */
		if (addr > 0x7F && addr < 0xFFFFFF80) {
			error("Branch address out of range!");
			return;
		}

		/* offset */
		putbyte(data_loccnt+1, addr);

		/* output line */
		println();
	}
}

#ifdef PCENGINE
/* ----
 * class3()
 * ----
 * 2 bytes, inherent addressing
 */

void
class3(int *ip)
{
	check_eol(ip);

	/* update location counter */
	loccnt += 2;

	/* generate code */
	if (pass == LAST_PASS) {
		/* opcodes */
		putbyte(data_loccnt, opval);
		putbyte(data_loccnt+1, optype);

		/* output line */
		println();
	}
}
#endif

/* ----
 * class4()
 * ----
 * various addressing modes
 */

void
class4(int *ip)
{
	int mode;
	int	i;

	/* get operand */
	mode = getoperand(ip, opflg, ';');
	if (!mode)
		return;

	/* make opcode */
	if (pass == LAST_PASS) {
		for (i = 0; i < 32; i++) {
			if (mode & (1 << i))
				break;
		}
		opval += opvaltab[optype][i];
	}

	/* generate code */
	switch(mode) {
	case ACC:
		/* one byte */
		if (pass == LAST_PASS)
			putbyte(data_loccnt, opval);

		loccnt++;
		break;

	case IMM:
	case ZP:
	case ZP_X:
	case ZP_Y:
	case ZP_IND:
	case ZP_IND_X:
	case ZP_IND_Y:
		/* two bytes */
		if (pass == LAST_PASS) {
			putbyte(data_loccnt, opval);
			putbyte(data_loccnt+1, value);
		}
		loccnt += 2;
		break;

	case ABS:
	case ABS_X:
	case ABS_Y:
	case ABS_IND:
	case ABS_IND_X:
		/* three bytes */
		if (pass == LAST_PASS) {
			putbyte(data_loccnt, opval);
			putword(data_loccnt+1, value);
		}
		loccnt += 3;
		break;
	}

	/* output line */
	if (pass == LAST_PASS)
		println();
}

#ifdef PCENGINE
/* ----
 * class5()
 * ----
 * 3 bytes, zp/relative addressing
 */

void
class5(int *ip)
{
	int	zp;
	int addr;
	int mode;

	/* update location counter */
	loccnt += 3;

	/* get first operand */
	mode = getoperand(ip, ZP, ',');
	zp   = value;
	if (!mode)
		return;

	/* get second operand */
	mode = getoperand(ip, ABS, ';');
	if (!mode)
		return;

	/* generate code */
	if (pass == LAST_PASS) {
		/* opcodes */
		putbyte(data_loccnt, opval);
		putbyte(data_loccnt+1, zp);

		/* calculate branch offset */
		addr = value - (loccnt + (page << 13));

		/* check range */
		if (addr > 0x7F && addr < 0xFFFFFF80) {
			error("Branch address out of range!");
			return;
		}

		/* offset */
		putbyte(data_loccnt+2, addr);

		/* output line */
		println();
	}
}

/* ----
 * class6()
 * ----
 * 7 bytes, src/dest/length
 */

void
class6(int *ip)
{
	int	i;
	int addr[3];

	/* update location counter */
	loccnt +=7;

	/* get operands */
    for (i = 0; i < 3; i++) {
		if (!evaluate(ip, (i < 2) ? ',' : ';'))
			return;
		if (pass == LAST_PASS) {
			if (value & 0xFFFF0000) {
				error("Operand size error!");
				return;
			}
		}
	    addr[i] = value;
	}

	/* generate code */
	if (pass == LAST_PASS) {
		/* opcodes */
		putbyte(data_loccnt, opval);
		putword(data_loccnt+1, addr[0]);
		putword(data_loccnt+3, addr[1]);
		putword(data_loccnt+5, addr[2]);

		/* output line */
		println();
	}
}

/* ----
 * class7()
 * ----
 * TST instruction
 */

void
class7(int *ip)
{
	int mode;
	int addr, imm;

	/* get first operand */
	mode = getoperand(ip, IMM, ',');
	imm  = value;
	if (!mode)
		return;

	/* get second operand */
	mode = getoperand(ip, (ZP | ZP_X | ABS | ABS_X), ';');
	addr = value;
	if (!mode)
		return;

	/* make opcode */
	if (mode & (ZP | ZP_X))
		opval = 0x83;
	if (mode & (ABS | ABS_X))
		opval = 0x93;
	if (mode & (ZP_X | ABS_X))
		opval+= 0x20;

	/* generate code */
	if (pass == LAST_PASS) {
		/* opcodes */
		putbyte(data_loccnt, opval);
		putbyte(data_loccnt+1, imm);

		if (mode & (ZP | ZP_X))
			/* zero page */
			putbyte(data_loccnt+2, addr);
		else
			/* absolute */
			putword(data_loccnt+2, addr);
	}

	/* update location counter */ 
	if (mode & (ZP | ZP_X))
		loccnt += 3;
	else
		loccnt += 4;

	/* output line */
	if (pass == LAST_PASS)
		println();
}

/* ----
 * class8()
 * ----
 * TAM/TMA instruction
 */

void
class8(int *ip)
{
	int mode;

	/* update location counter */
	loccnt += 2;

	/* get operand */
	mode = getoperand(ip, IMM, ';');
	if (!mode)
		return;

	/* generate code */
	if (pass == LAST_PASS) {
		/* check page index */
		if (value & 0xF8) {
			error("Incorrect page index!");
			return;
		}

		/* opcodes */
		putbyte(data_loccnt, opval);
		putbyte(data_loccnt+1, (1 << value));

		/* output line */
		println();
	}
}

/* ----
 * class9()
 * ----
 * RMB/SMB instructions
 */

void
class9(int *ip)
{
	int bit;
	int mode;

	/* update location counter */
	loccnt += 2;

	/* get the bit index */
	mode = getoperand(ip, IMM, ',');
	bit  = value;
	if (!mode)
		return;

	/* get the zero page address */
	mode = getoperand(ip, ZP, ';');
	if (!mode)
		return;

	/* generate code */
	if (pass == LAST_PASS) {
		/* check bit number */
		if (bit > 7) {
			error("Incorrect bit number!");
			return;
		}

		/* opcodes */
		putbyte(data_loccnt, opval + (bit << 4));
		putbyte(data_loccnt+1, value);

		/* output line */
		println();
	}
}

/* ----
 * class10()
 * ----
 * BBR/BBS instructions
 */

void
class10(int *ip)
{
	int bit;
	int zp;
	int mode;
	int addr;

	/* update location counter */
	loccnt += 3;

	/* get the bit index */
	mode = getoperand(ip, IMM, ',');
	bit  = value;
	if (!mode)
		return;

	/* get the zero page address */
	mode = getoperand(ip, ZP, ',');
	zp   = value;
	if (!mode)
		return;

	/* get the jump address */
	mode = getoperand(ip, ABS, ';');
	if (!mode)
		return;

	/* generate code */
	if (pass == LAST_PASS) {
		/* check bit number */
		if (bit > 7) {
			error("Incorrect bit number!");
			return;
		}

		/* opcodes */
		putbyte(data_loccnt, opval + (bit << 4));
		putbyte(data_loccnt+1, zp);

		/* calculate branch offset */
		addr = value - (loccnt + (page << 13));

		/* check range */
		if (addr > 0x7F && addr < 0xFFFFFF80) {
			error("Branch address out of range!");
			return;
		}

		/* offset */
		putbyte(data_loccnt+2, addr);

		/* output line */
		println();
	}
}
#endif

/* ----
 * getoperand()
 * ----
 */

int
getoperand(int *ip, int flag, int last_char)
{
	char c;
	int code;
	int mode;
	int pos;
	int end;

	/* skip spaces */
	while (isspace(prlnbuf[*ip]))
		(*ip)++;

	/* check addressing mode */
	switch (prlnbuf[*ip]) {
	case '\0':
	case ';':
		/* no operand */
		error("Operand missing!");
		return (0);

	case 'A':
	case 'a':
		/* accumulator */
		c = prlnbuf[(*ip)+1];
		if (isspace(c) || c == '\0' || c == ';' || c == ',') {
			mode = ACC;
			(*ip)++;
			break;
		}

	default:
		/* other */
		switch(prlnbuf[*ip]) {
		case '#':
			/* immediate */
			mode = IMM;
			(*ip)++;
			break;

		case '<':
			/* zero page */
			mode = ZP | ZP_X | ZP_Y;
			(*ip)++;
			break;

		case '[':
			/* indirect */
			mode = ABS_IND | ABS_IND_X | ZP_IND | ZP_IND_X | ZP_IND_Y;
			(*ip)++;
			break;

		default:
			/* absolute */
			mode = ABS | ABS_X | ABS_Y;
			break;
		}

		/* get value */
		if (!evaluate(ip, 0))
			return (0);

		/* check value on last pass */
		if (pass == LAST_PASS) {
			if (mode & (ZP | ZP_X | ZP_Y | ZP_IND | ZP_IND_X | ZP_IND_Y) & flag) {
				if ((value & 0xFFFFFF00) && ((value & 0xFFFFFF00) != RAM_BASE)) {
					error("Incorrect zero page address!");
					return (0);
				}
			}
			else if (mode & (IMM) & flag) {
				if ((value > 0xFF) && (value < 0xFFFFFF00)) {
					error("Incorrect immediate value!");
					return (0);
				}
			}
			else if (mode & (ABS | ABS_X | ABS_Y | ABS_IND | ABS_IND_X) & flag) {
				if (value & 0xFFFF0000) {
					error("Incorrect absolute address!");
					return (0);
				}
			}
		}

		/* check addressing mode */
		code = 0;
		end = 0;
		pos = 0;

		while (!end) {
			c = prlnbuf[*ip];
			if (c == ';' || c == '\0')
				break;
			switch (toupper(c)) {
			case ',':		/* , = 4 */
				if (!pos)
					 pos = *ip;
				else {
					end = 1;
					break;
				}
				code++;
			case ']':		/* ] = 3 */
				code++;
			case 'X':		/* X = 2 */
				code++;
			case 'Y':		/* Y = 1 */
				code++;
				code <<= 4;
			case ' ':
			case '\t':
				(*ip)++;
				break;
			default:
				code = 0xFFFF;
				end  = 1;
				break;
			}
		}

		switch(code) {
		case 0x0000: mode &= (ABS | ZP | IMM); break;		//
		case 0x0030: mode &= (ZP_IND | ABS_IND); break;		// ]
		case 0x0410: mode &= (ABS_Y | ZP_Y); break;			// ,Y
		case 0x0420: mode &= (ABS_X | ZP_X); break;			// ,X
		case 0x4230: mode &= (ZP_IND_X | ABS_IND_X); break;	// ,X]
		case 0x3410: mode &= (ZP_IND_Y); break;				// ],Y
		default:
			mode &= (ABS | ZP | IMM);
			if (pos)
				*ip = pos;
			break;
		}
		break;
	}

	/* compare addressing mode */
	mode &= flag;
	if (!mode) {
		error("Incorrect addressing mode!");
		return (0);
	}

	/* skip spaces */
	while (isspace(prlnbuf[*ip]))
		(*ip)++;

	/* get last char */
	c = prlnbuf[*ip];

	/* check if it's what the user asked for */
	switch (last_char) {
	case ';':
		/* last operand */
		if (c != ';' && c != '\0') {
			error("Syntax error!");
			return (0);
		}
		(*ip)++;
		break;		

	case ',':
		/* need more operands */
		if (c != ',') {
			error("Argument missing!");
			return (0);
		}
		(*ip)++;
		break;		
	}

	/* ok */
	return (mode);
}

/* ----
 * getstring()
 * ----
 * get a string from prlnbuf
 */

int
getstring(int *ip, char *buffer, int size)
{
	char c;
	int i;

	/* skip spaces */
	while (isspace(prlnbuf[*ip]))
		(*ip)++;

	/* string must be enclosed */
	if (prlnbuf[(*ip)++] != '\"') {
		error("Incorrect string!");
		return (0);
	}

	/* get string */
	i = 0;
	for (;;) {
		c = prlnbuf[(*ip)++];
		if (c == '\"')
			break;
		if (i >= size) {
			error("String too long!");
			return (0);
		}
		buffer[i++] = c;
	}

	/* end the string */
	buffer[i] = '\0';

	/* skip spaces */
	while (isspace(prlnbuf[*ip]))
		(*ip)++;

	/* ok */
	return (1);
}

