#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "externs.h"
#include "protos.h"

char pseudo_flag[] = {
	0x0C, 0x0C, 0x0F, 0x0F, 0x0F, 0x0C, 0x0C, 0x0C, 0x0F, 0x0C,
	0x0C, 0x0C, 0x0C, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F,
	0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0C,
	0x0F, 0x0F, 0x0F, 0x0C, 0x0C, 0x0C, 0x0C, 0x0F, 0x0F, 0x0F,
	0x0F
};

/* ----
 * pseudo()
 * ----
 * pseudo operations processor
 */

void
pseudo(int *ip)
{
	/* check if the directive is allowed in the current section */
	if (~pseudo_flag[opval] & (1 << section))
		fatal_error("Directive not allowed in the current section!");

	/* execute directive */
	switch(opval) {
	case P_DB:		do_db(ip); break;
	case P_DW:		do_dw(ip); break;
	case P_EQU:		do_equ(ip); break;
	case P_PAGE:	do_page(ip); break;
	case P_ORG:		do_org(ip); break;
	case P_BANK:	do_bank(ip); break;
	case P_INCBIN:	do_incbin(ip); break;
	case P_INCLUDE:	do_include(ip); break;

#ifdef PCENGINE
	case P_INCSPR:
	case P_INCPAL:
	case P_INCBAT:
#endif
	case P_INCCHR:	do_pcx(ip); break;
	case P_MACRO:	do_macro(ip); break;
	case P_ENDM:	do_endm(ip); break;
	case P_LIST:	check_eol(ip); clist = 1; xlist = 1; break;
	case P_MLIST:	check_eol(ip); mlist = 1; break;
	case P_NOLIST:	check_eol(ip); clist = 0; break;
	case P_NOMLIST:	check_eol(ip); mlist = mlist_opt; break;
	case P_RSSET:	do_rsset(ip); break;
	case P_RS:		do_rs(ip); break;
	case P_DS:		do_ds(ip); break;
	case P_IF:		do_if(ip); break;
	case P_ELSE:	do_else(ip); break;
	case P_ENDIF:	do_endif(ip); break;
	case P_FAIL:	do_fail(ip); break;
	case P_ZP:
	case P_BSS:
	case P_CODE:
	case P_DATA:	do_section(ip); break;
	case P_FUNC:	do_func(ip); break;
	case P_IFDEF:
	case P_IFNDEF:	do_ifdef(ip); break;
	case P_DEFCHR:	do_defchr(ip); break;

#ifdef PCENGINE
	case P_DEFSPR:	do_defspr(ip); break;
	case P_DEFPAL:	do_defpal(ip); break;
	case P_VRAM:	do_vram(ip); break;
	case P_PAL:		do_pal(ip); break;
#endif

#ifdef NES
	case P_INESPRG:	do_inesprg(ip); break;
	case P_INESCHR:	do_ineschr(ip); break;
	case P_INESMAP:	do_inesmap(ip); break;
	case P_INESMIR:	do_inesmir(ip); break;
#endif
	}

	switch (opval) {
	case P_VRAM:
	case P_PAL:
		break;

	default:
		lastlabl = NULL;
		break;
	}
}

/* .db pseudo */

void
do_db(int *ip)
{
	unsigned char c;

	/* define label */
	labldef(loccnt, 1);

	/* output infos */
	data_loccnt = loccnt;
	data_level  = 2;

	/* skip spaces */
	while (isspace(prlnbuf[++(*ip)]));

	/* get bytes */
	for (;;) {
		/* ASCII string */
		if (prlnbuf[*ip] == '\"') {
			for (;;) {
				c = prlnbuf[++(*ip)];
				if (c == '\"')
					break;
				if (c == '\0') {
					error("Unterminated ASCII string!");
					return;
				}
				if (c == '\\') {
					c = prlnbuf[++(*ip)];
					switch(c) {
					case 'r':
						c = '\r';
						break;
					case 'n':
						c = '\n';
						break;
					case 't':
						c = '\t';
						break;
					}
				}
				/* store char on last pass */
				if (pass == LAST_PASS)
					putbyte(loccnt, c);

				/* update location counter */
				loccnt++;
			}
			(*ip)++;
		}
		/* bytes */
		else {
			/* get a byte */
			if (!evaluate(ip, 0))
				return;

			/* update location counter */
			loccnt++;

			/* store byte on last pass */
			if (pass == LAST_PASS) {
				/* check for overflow */
				if ((value > 0xFF) && (value < 0xFFFFFF80)) {
					error("Overflow error!");
					return;
				}

				/* store byte */
				putbyte(loccnt-1, value);
			}
		}

		/* check if there's another byte */
		c = prlnbuf[(*ip)++];

		if (c != ',')
			break;
	}

	/* check error */
	if (c != ';' && c != '\0') {
		error("Syntax error!");
		return;
	}

	/* output line */
	if (pass == LAST_PASS)
		println();
}

/* .dw pseudo */

void
do_dw(int *ip)
{
	char c;

	/* define label */
	labldef(loccnt, 1);

	/* output infos */
	data_loccnt = loccnt;
	data_format = 2;
	data_level  = 2;

	/* get data */
	for (;;) {
		/* get a word */
		if (!evaluate(ip, 0))
			return;

		/* update location counter */
		loccnt += 2;

		/* store word on last pass */
		if (pass == LAST_PASS) {
			/* check for overflow */
			if ((value > 0xFFFF) && (value < 0xFFFF8000)) {
				error("Overflow error!");
				return;
			}

			/* store word */
			putword(loccnt-2, value);
		}

		/* check if there's another word */
		c = prlnbuf[(*ip)++];

		if (c != ',')
			break;
	}

	/* check error */
	if (c != ';' && c != '\0') {
		error("Syntax error!");
		return;
	}

	/* output line */
	if (pass == LAST_PASS)
		println();
}

/* .equ pseudo */

void
do_equ(int *ip)
{
	/* get value */
	if (!evaluate(ip, ';'))
		return;

	/* assign value to the label */
	labldef(value, 0);

	/* output line */
	if (pass == LAST_PASS) {
		loadlc(value, 1);
		println();
	}
}

/* .page pseudo */

void
do_page(int *ip)
{
	/* define label */
	labldef(loccnt, 1);

	/* get page address */
	if (!evaluate(ip, ';'))
		return;
	if (value & 0xFFFF1FFF) {
		error("Invalid page address!");
		return;
	}
	page = (value >> 13);

	/* output line */
	if (pass == LAST_PASS) {
		loadlc(value, 1);
		println();
	}
}

/* .org pseudo */

void
do_org(int *ip)
{
	/* get the .org value */
	if (!evaluate(ip, ';'))
		return;

	/* check for undefined symbol - they are not allowed in .org */
	if (undef != 0) {
		error("Undefined symbol in operand field!");
		return;
	}

	/* section switch */
	switch (section) {
	case S_ZP:
		/* zero page section */
		if ((value & 0xFFFFFF00) && ((value & 0xFFFFFF00) != RAM_BASE)) {
			error("Invalid address!");
			return;
		}
		break;

	case S_BSS:
		/* ram section */
		if ((value < RAM_BASE) || (value >= (RAM_BASE + RAM_LIMIT))) {
			error("Invalid address!");
			return;
		}
		break;

	case S_CODE:
	case S_DATA:
		/* code and data section */
		if (value & 0xFFFF0000) {
			error("Invalid address!");
			return;
		}
		page = (value >> 13) & 0x07;
		break;
	}

	/* set location counter */
	loccnt = (value & 0x1FFF);

	/* set label value if there was one */
	labldef(loccnt, 1);

	/* output line on last pass */
	if (pass == LAST_PASS) {
		loadlc(value, 1);
		println();
	}
}

/* .bank pseudo */

void
do_bank(int *ip)
{
	char name[128];

	/* define label */
	labldef(loccnt, 1);

	/* get bank index */
	if (!evaluate(ip, 0))
		return;
	if (value > 0x7F) {
		error("Bank index out of range!");
		return;
	}

	/* check if there's a bank name */
	switch (prlnbuf[*ip]) {
	case ';':
	case '\0':
		break;
		
	case ',':
		/* get name */
		(*ip)++;
		if (!getstring(ip, name, 127))
			return;

		/* check name validity */
		if (strlen(bank_name[value])) {
		#ifndef LINUX
			if (stricmp(name, bank_name[value])) {
		#else
			if (strcasecmp(name, bank_name[value])) {
		#endif
				error("Different bank names not allowed!");
				return;
			}
		}

		/* copy name */			
		strcpy(bank_name[value], name);

		/* check end of line */
		if (!check_eol(ip))
			return;

		/* ok */
		break;

	default:
		error("Syntax error!");
		return;
	}

	/* backup current bank infos */
	bank_glabl[section][bank]  = glablptr;
	bank_loccnt[section][bank] = loccnt;
	bank_page[section][bank]   = page;

	/* get new bank infos */
	bank     = value;
	page     = bank_page[section][bank];
	loccnt   = bank_loccnt[section][bank];
	glablptr = bank_glabl[section][bank];

	/* output on last pass */
	if (pass == LAST_PASS) {
		loadlc(bank, 1);
		println();
	}
}

/* .incbin pseudo */

void
do_incbin(int *ip)
{
	FILE *f;
	char name[128];
	int size;

	/* define label */
	labldef(loccnt, 1);

	/* get file name */
	if (!getstring(ip, name, 127))
		return;

	/* open file */
	if ((f = open_file(name, "rb")) == NULL) {
		error("Can not open file!");
		return;
	}

	/* get file size */
	fseek(f, 0, SEEK_END);
	size = ftell(f);
	fseek(f, 0, SEEK_SET);

	/* check if it will fit in the rom */
	if (((bank << 13) + loccnt + size) > (8192 * 128)) {
		fclose(f);
		error("File too big!");
		return;
	}

	/* load data on last pass */
	if (pass == LAST_PASS) {
		fread(&rom[bank][loccnt], 1, size, f);
		memset(&map[bank][loccnt], section + (page << 5), size);

		/* output line */
		loadlc(loccnt, 0);
		println();
	}

	/* close file */
	fclose(f);

	/* update bank and location counters */
	bank  += (loccnt + size) >> 13;
	loccnt = (loccnt + size) & 0x1FFF;
	if (bank > max_bank) {
		if (loccnt)
			max_bank = bank;
		else
			max_bank = bank - 1;
	}
}

/* .include pseudo */

void
do_include(int *ip)
{
	char name[128];

	/* define label */
	labldef(loccnt, 1);

	/* get file name */
	if (!getstring(ip, name, 127))
		return;

	/* open file */
	if (open_input(name) == -1) {
		error("Can not open file!");
		return;
	}

	/* output line */
	if (pass == LAST_PASS)
		println();
}

/* .rsset pseudo */

void
do_rsset(int *ip)
{
	/* define label */
	labldef(loccnt, 1);

	/* get 'rs' counter value */
	if (!evaluate(ip, ';'))
		return;
	if (value & 0xFFFF0000) {
		error("Address out of range!");
		return;
	}
	rsbase = value;

	/* output line */
	if (pass == LAST_PASS) {
		loadlc(rsbase, 1);
		println();
	}
}

/* .rs pseudo */

void
do_rs(int *ip)
{
	/* define label */
	labldef(rsbase, 0);

	/* get the number of bytes to reserve */
	if (!evaluate(ip, ';'))
		return;

	/* ouput line */
	if (pass == LAST_PASS) {
		loadlc(rsbase, 1);
		println();
	}

	/* update 'rs' counter */
	rsbase += value;
	if (rsbase & 0xFFFF0000)
		error("Address out of range!");
}

/* .ds pseudo */

void
do_ds(int *ip)
{
	int limit = 0;
	int addr;

	/* define label */
	labldef(loccnt, 1);

	/* get the number of bytes to reserve */
	if (!evaluate(ip, ';'))
		return;

	/* section switch */
	switch (section) {
	case S_ZP:
		/* zero page section */
		limit = ZP_LIMIT;
		break;	

	case S_BSS:
		/* ram section */
		limit = RAM_LIMIT;
		break;

	case S_CODE:
	case S_DATA:
		/* code and data sections */
		limit = 0x2000;
		break;
	}

	/* check range */
	if ((loccnt + value) > limit) {
		error("Out of range!");
		return;
	}

	/* update max counter for zp and bss sections */
	addr = loccnt + value;

	switch (section) {
	case S_ZP:
		/* zero page */
		if (addr > max_zp)
			max_zp = addr;
		break;

	case S_BSS:
		/* ram page */
		if (addr > max_bss)
			max_bss = addr;
		break;
	}

	/* output line on last pass */
	if (pass == LAST_PASS) {
		switch (section) {
		case S_CODE:
		case S_DATA:
			memset(&rom[bank][loccnt], 0, value);
			memset(&map[bank][loccnt], section + (page << 5), value);
			if (bank > max_bank)
				max_bank = bank;
			break;
		}
		loadlc(loccnt + (page << 13), 1);
		println();
	}

	/* update location counter */
	loccnt += value;
}

/* .fail pseudo */

void
do_fail(int *ip)
{
	fatal_error("Compilation failed!");
}

/* .zp/.bss/.code/.data pseudo */

void
do_section(int *ip)
{
	if (section != optype) {
		/* backup current section data */
		section_bank[section] = bank;
		bank_glabl[section][bank] = glablptr;
		bank_loccnt[section][bank] = loccnt;
		bank_page[section][bank] = page;

		/* change section */
		section = optype;

		/* switch to the new section */
		bank = section_bank[section];
		page = bank_page[section][bank];
		loccnt = bank_loccnt[section][bank];
		glablptr = bank_glabl[section][bank];
	}

	/* output line */
	if (pass == LAST_PASS) {
		loadlc(loccnt + (page << 13), 1);
		println();
	}
}

/* .incchr/spr/pal/bat pseudo */

void
do_pcx(int *ip)
{
	char name[128];
	int  addr;
	char c;
	int  arg[5];
	int  n = 0;

	/* define label */
	labldef(loccnt, 1);

	/* output infos */
	addr = loccnt;

	/* get PCX file name */
	if (!getstring(ip, name, 127))
		return;

	/* get args */
	for (;;) {
		while (isspace(c = prlnbuf[(*ip)++]));
		if ((c != ',') && (c != ';') && (c != 0)) {
			error("Syntax error!");
			return;
		}
		if (c != ',')
			break;
		if (!evaluate(ip, 0))
			return;
		arg[n++] = value;
		if (n == 5)
			break;
	}			

	/* check number of args */
	if (optype & (1 << n)) {
		error("Invalid number of arguments!");
		return;
	}

	/* load and unpack the PCX */
	if (!load_pcx(name))
		return;

	/* dispatch PCX directives */
	switch (opval) {
	case P_INCCHR:
		do_incchr(n, arg);
		break;

#ifdef PCENGINE
	case P_INCSPR:
		do_incspr(n, arg);
		break;
	case P_INCPAL:
		do_incpal(n, arg);
		break;
	case P_INCBAT:
		do_incbat(n, arg);
		break;
#endif
	}

	/* output */
	if (pass == LAST_PASS) {
		loadlc(addr, 0);
		println();
	}
}

/* convert a PCX to characters  */

void
do_incchr(int n, unsigned int *arg)
{
	unsigned char *data;
	unsigned int i, j, x, y, w, h;
	unsigned int size;

	/* get args */
	if (!check_args(arg, n, &x, &y, &w, &h, 8))
		return;

	/* clear the tile plane buffer */
	memset(tile_plane, 0, 64);

	/* make the data */
	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++) {
			/* tile address in the PCX */
			data = pcx_buf + (x + (j << 3)) + ((y + (i << 3)) * pcx_w);

			/* encode the tile */
			size = pack_8x8_tile(data, CHUNKY_TILE, pcx_w);

			if (!size)
				return;

			/* store tile */
			putbuffer(tile_plane, size);
		}
	}
}

