#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "externs.h"
#include "protos.h"

int    infile_error;
int    infile_num;
char   i_file[128];
FILE  *iptr;
struct t_input_info input_file[8];
char   incpath[10][128];

/* init the include path */

void
init_path(void)
{
	char	*p,*pl;
	int	i, l;

	p = getenv(INC_ENV);

	if (p == NULL)
		return;

	for (i = 0; i < 10; i++) {

		pl = strchr(p, ';');

		if (pl == NULL)
			l = strlen(p);
		else
			l = pl-p;

		if (l == 0) {
			incpath[i][0] = '\0';
		} else {
			strncpy(incpath[i],p,l);
			p += l;
			while (*p == ';') p++;
		}

	  
	        #ifndef LINUX
		if (incpath[i][strlen(incpath[i])] != '\\') {
			strcat(incpath[i], "\\");
		}
		#else
		if (incpath[i][strlen(incpath[i])] != '/') 
		    {
		      strcat(incpath[i], "/");
		    }
		#endif
		
	}
}

/*
 * readline :
 * ---------
 * reads and formats an input line.
 *
 */

int
readline(void)
{
	char *ptr, *arg, num[8];
	int j, n;
	int	i;		/* pointer into prlnbuf */
	int	c;		/* current character		*/
	int	temp;	/* temp used for line number conversion */

  #ifdef TEST
    fprintf(stderr,"Entering readline\n");
    #endif
  
start:
	for (i = 0; i < LAST_CH_POS; i++)
		prlnbuf[i] = ' ';
#ifdef TEST
  fprintf(stderr,"will expand macro\n");
  #endif
	/* if 'expand_macro' is set get a line from macro buffer instead */
	if (expand_macro) {
		if (mlptr == NULL) {
			while (mlptr == NULL) {
				midx--;
				mlptr = mstack[midx];
				mcounter = mcntstack[midx];
				if (midx == 0) {
					mlptr = NULL;
					expand_macro = 0;
					break;
				}
			}
		}

		/* expand line */
		if (mlptr) {
			i = SFIELD;
			ptr = mlptr->data;
			for (;;) {
				c = *ptr++;
				if (c == '\0')
					break;
				if (c != '\\')
					prlnbuf[i++] = c;
				else {
					c = *ptr++;
					prlnbuf[i] = '\0';

					/* \@ */
					if (c == '@') {
						n = 4;
						sprintf(num, "%04i", mcounter);
						arg = num;
					}

					/* \# */
					else if (c == '#') {
						for (j = 9; j > 0; j--)
							if (strlen(marg[midx][j - 1]))
								break;
						n = 1;
						sprintf(num, "%i", j);
						arg = num;
					}

					/* \?1 - \?9 */
					else if (c == '?') {
						c = *ptr++;
						if (c >= '1' && c <= '9') {
							n = 1;
							sprintf(num, "%i", macro_getargtype(marg[midx][c - '1']));
							arg = num;
						}
						else {
							error("Invalid macro argument index!");
							return (-1);
						}
					}

					/* \1 - \9 */
					else if (c >= '1' && c <= '9') {
						j   = c - '1';
						n   = strlen(marg[midx][j]);
						arg = marg[midx][j];
					}

					/* unknown macro special command */
					else {
						error("Invalid macro argument index!");
						return (-1);
					}

					/* check for line overflow */
					if ((i + n) >= LAST_CH_POS - 1) {
						error("Invalid line length!");
						return (-1);
					}

					/* copy macro string */
					strncpy(&prlnbuf[i], arg, n);
					i += n;
				}
				if (i >= LAST_CH_POS - 1)
					i  = LAST_CH_POS - 1;
			}
			prlnbuf[i] = '\0';
			mlptr = mlptr->next;
			return (0);
		}
	}

  #ifdef TEST
    fprintf(stderr,"macro expanded\n");
    #endif
  
	/* put source line number into prlnbuf */
	i = 4;
	temp = ++slnum;
	while (temp != 0) {
		prlnbuf[i--] = temp % 10 + '0';
		temp /= 10;
	}

  
  #ifdef TEST
    fprintf(stderr,"will get a line\n");
    #endif
	/* get a line */
	i = SFIELD;
	c = getc(iptr);
	if (c == EOF) {
		if (close_input())
			return (-1);
		goto start;
	}
  #ifdef TEST
    fprintf(stderr,"got a line\n");
    #endif
	for (;;) {
	  #ifdef TEST
	    fprintf(stderr,"checking : %d, fin de ligne = %d\n",c,((c=='\n')||(c==EOF)||(c==13)));
	    #endif
		/* check for the end of line */
		if(c == '\r')
	            {
		      c = getc(iptr);
		      continue;
		    }
	  
		if((c == '\n') || (c == EOF))
	    {
	      #ifdef TEST
		fprintf(stderr,"VICTORY!!!!\n");
		#endif
			break;
	    }
	  
		/* store char in the line buffer */
		prlnbuf[i] = c;
		i += (i < LAST_CH_POS) ? 1 : 0;

		/* expand tab char to space */
		if (c == '\t') {
			prlnbuf[--i] = ' ';
			i += (8 - ((i - SFIELD) % 8));
		}

		/* get next char */
		c = getc(iptr);
	}
  
  #ifdef TEST
    fprintf(stderr,"end of line got\n");
    #endif
	prlnbuf[i] = '\0';
	return(0);
}

/*
 * open_input :
 * -----------
 * open an input file, up to 7 levels.
 *
 */

int
open_input(char *name)
{
	FILE *fptr;
	char *p;
	int i;

	if (infile_num == 7) {
		error("Too many include levels!");
		return (1);
	}
	if (infile_num) {
		input_file[infile_num].lnum = slnum;
		input_file[infile_num].fptr = iptr;
	}				
	strcpy(i_file, name);
    
        #ifndef LINUX
	strupr(i_file);
        #endif
    
	if ((p = strrchr(i_file, '.')) != NULL) {
		if (strchr(p, '\\'))
			strcat(i_file, ".ASM");
	} else 
		strcat(i_file, ".ASM");
	if (infile_num) {
		for (i = 1; i < infile_num; i++) {
			if (!strcmp(input_file[i].name, i_file)) {
				error("Repeated include file!");
				return (1);
			}
		}
	}				
	if ((fptr = open_file(i_file, "r")) == NULL)
		return (-1);

	iptr = fptr;
	slnum = 0;
	infile_num++;
	input_file[infile_num].fptr = fptr;
	strcpy(input_file[infile_num].name, i_file);
	if ((pass == LAST_PASS) && (xlist) && (list_level))
		fprintf(lptr, "#[%i]   %s\n", infile_num, input_file[infile_num].name);

	/* ok */
	return (0);
}

/*
 * close_input :
 * ------------
 * close an input file, return -1 if no more files in the stack.
 *
 */

int
close_input(void)
{
	if (in_macro) {
		fatal_error("Incomplete MACRO definition!");
		return (-1);
	}
	if (in_if) {
		fatal_error("Incomplete IF/ENDIF statement!");
		return (-1);
	}
	if (infile_num <= 1)
		return (-1);

	fclose(iptr);
	infile_num--;
	infile_error = -1;
	slnum = input_file[infile_num].lnum;
	iptr  = input_file[infile_num].fptr;
	if ((pass == LAST_PASS) && (xlist) && (list_level))
		fprintf(lptr, "#[%i]   %s\n", infile_num, input_file[infile_num].name);

	/* ok */
	return (0);
}

/* open a file */

FILE *
open_file(char *name, char *mode)
{
	FILE 	*fileptr;
	char	testname[256];
	int	i;

	fileptr = fopen(name, mode);
	if (fileptr != NULL) return(fileptr);

	for (i = 0; i < 10; i++) {
		if (strlen(incpath[i])) {
			strcpy(testname, incpath[i]);
			strcat(testname, name);
	
			fileptr = fopen(testname, mode);
			if (fileptr != NULL) break;
		}
	}

	return (fileptr);
}

