/*  PC-Engine and NES assembler
 *  ----
 *  This program was originaly a 6502 assembler written by J. H. Van Ornum,
 *  it has been modified and enhanced to support the PC Engine and NES consoles.
 *
 *  This program is freeware. You are free to distribute, use and modifiy it
 *  as you wish.
 *
 *  Enjoy!
 *  ----
 *  Original 6502 version by:
 *    J. H. Van Ornum
 *
 *  PC-Engine version by:
 *    David Michel
 *    Dave Shadoff
 *
 *  NES version by:
 *    Charles Doty
 * 
 *  LINUX adaption by :
 *    Olivier Jolly
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "externs.h"
#include "protos.h"
#include "vars.h"
#include "inst.h"

char  o_file[128];
char  l_file[128];
FILE *lptr;
unsigned char section_name[4][8] = { "  ZP", " BSS", "CODE", "DATA" };
int dump_seg;
int rom_used;
int rom_free;
int header_opt;

/* ----
 * main()
 * ----
 */

int
main(int argc, char **argv)
{
	FILE *f;
	char *p;
	int i, j;
	int file = 0;

	/* init variables */
	list_level = 2;
	mlist_opt = 0;
	header_opt = 1;

	/* display assembler version */
        printf(ASM_TITLE"\n\n");

	/* parse command line */
	if (argc > 1) {
		for (i = 1; i < argc; i++) {
			if (argv[i][0] == '-') {
				/* segment usage */
				if (!strcmp(argv[i], "-s"))
					dump_seg = 1;
				else if (!strcmp(argv[i], "-S"))
					dump_seg = 2;

				/* forces macros expansion */
				else if (!strcmp(argv[i], "-m"))
					mlist_opt = 1;

				/* no header */
				else if (!strcmp(argv[i], "-raw"))
					header_opt = 0;

				/* output level */
				else if (!strncmp(argv[i], "-l", 2)) {
					if (strlen(argv[i]) == 2)
						list_level = atol(argv[++i]);
					else
						list_level = atol(&argv[i][2]);

					/* check range */
					if (list_level < 0 || list_level > 3)
						list_level = 2;
				}

				/* help */
				else if (!strcmp(argv[i], "-?")) {
			        printf(ASM_NAME" [-options] [-? (For Help)] infile\n\n");
					printf("-s/S   : Shows segment usage.\n");
					printf("-l #   : Listing file output level (0-3).\n");
					printf("-raw   : Prevents adding ROM header.\n");
					printf("-m     : Forces macros expansion in listing.\n");
					printf("infile : File to be assembled.\n");
					return (0);
				}
			}
			else {
				strcpy(i_file, argv[i]);
			        #ifndef LINUX
				strupr(i_file);
			        #endif
				file++;
			}
		}
	}
	if (!file) {
        printf(ASM_NAME" [-options] [-? (For Help)] infile\n\n");
		return (0);
	}

#ifdef TEST
  fprintf(stderr,"Will add file extension\n");
#endif
  
	/* auto-add file extensions */
	if ((p = strrchr(i_file, '.')) != NULL) {
		if (!strchr(p, '\\'))
		   *p = '\0';
		else
			p = NULL;
	}
	strcpy(o_file, i_file);
	strcpy(l_file, i_file);
	strcat(o_file, ROM_EXT);
	strcat(l_file, ".LST");

	if (p)
	   *p = '.';
	else
           strcat(i_file, ".ASM");

#ifdef TEST
  
  fprintf(stderr,"Will include path\n");
  
  #endif
  
	/* init include path */
	init_path();

	/* open the input file */
	if (open_input(i_file)) {
		fprintf(stdout, "Can not open input file '%s'!\n", i_file);
		exit(1);
	}

	/* clear the ROM array */
	memset(rom,    0, 8192 * 128);
	memset(map, 0xFF, 8192 * 128);

	/* clear symbol hash tables */
	for (i = 0; i < 256; i++) {
		hash_tbl[i]  = NULL;
		macro_tbl[i] = NULL;
		func_tbl[i]  = NULL;
		inst_tbl[i]  = NULL;
	}

	/* fill the instruction hash table */
	addinst(base_inst);
	addinst(base_pseudo);

#ifdef TEST
  
  fprintf(stderr,"init machine\n");
  
  #endif
  
	/* machine initialization */
	init_machine();

 #ifdef TEST
    
    fprintf(stderr,"Machine initialised\n");
    
    #endif
  
	/* assemble */
	for (pass = FIRST_PASS; pass <= LAST_PASS; pass++) {
	  #ifdef TEST
	    
	    fprintf(stderr,"PASS : %d\n",pass);
	    
	    #endif
		infile_error = -1;
		errcnt = 0;
		page = 7;
		bank = 0;
		max_zp = 0x00;
		max_bss = 0x0200;
		max_bank = 0;
		loccnt = 0;
		slnum = 0;
		mcounter = 0;
		mcntmax = 0;
		clist = 0;
		mlist = mlist_opt;
		glablptr = NULL;
		skip_lines = 0;
		rsbase = 0;

		/* init predefined symbol(s) in first pass */
		if (pass == FIRST_PASS) {
			/* MAGICKIT = 1 */
			strcpy(symbol, "\x08MAGICKIT");
			lablptr = stlook(1);
			labldef(1, 0);
		}

	  #ifdef TEST
	    fprintf(stderr,"Init banks\n");
	    #endif
	  
		/* initialize banks */
		for (i = 0; i < 4; i++) {
			for (j = 0; j < 128; j++) {
				bank_loccnt[i][j] = 0;
				bank_glabl[i][j] = NULL;
				bank_page[i][j] = 0;
			}
		}

		/* initialize sections */
		section_bank[S_ZP]        = 0xF8;
		bank_page[S_ZP][0xF8]     = RAM_PAGE;
		bank_loccnt[S_ZP][0xF8]   = 0x0000;
		section_bank[S_BSS]       = 0xF8;
		bank_page[S_BSS][0xF8]    = RAM_PAGE;
		bank_loccnt[S_BSS][0xF8]  = 0x0200;
		section_bank[S_CODE]      = 0x00;
		bank_page[S_CODE][0x00]   = 0x07;
		bank_loccnt[S_CODE][0x00] = 0x0000;
		section_bank[S_DATA]      = 0x00;
		bank_page[S_DATA][0x00]   = 0x07;
		bank_loccnt[S_DATA][0x00] = 0x0000;
		section = S_CODE;
#ifdef TEST
	  fprintf(stderr,"Will readline");
	  #endif
		/* assemble */
		while (readline() != -1) {
		  #ifdef TEST
		    fprintf(stderr,"Line read : %d\n",loccnt);
		    #endif
			assemble();
			if (loccnt > 0x2000) {
				error("Out of range, bank offset > $1FFF!");
				break;
			}
			if (stop_pass)
				break;
		}
	  #ifdef TEST
	    fprintf(stderr,"assembling finished\n");
	    #endif

		/* abord pass on errors */
		if (errcnt) {
			fprintf(stdout, "# %d ERROR(s)\n", errcnt);
			break;
		}

		/* rewind input file */
		rewind(iptr);

		/* open the listing file at the end of the first pass */
		if (pass == FIRST_PASS) {
			if (xlist && list_level) {
				if ((lptr = fopen(l_file, "w")) == NULL) {
					fprintf(stdout, "Can not open listing file '%s'!\n", l_file);
					exit(1);
				}
				fprintf(lptr, "#[1]   %s\n", input_file[1].name);
			}
		}
	}

	/* write rom */
	if (errcnt == 0) {
		/* open rom binary file */
		if ((f = fopen(o_file, "wb")) == NULL) {
			fprintf(stdout, "Can not open object file '%s'!\n", o_file);
			exit(1);
		}

		/* header */
		if (header_opt)
			write_header(f, max_bank + 1);

		/* rom */
		fwrite(rom, 8192, (max_bank + 1), f);
		fclose(f);
	}

  #ifdef TEST
    fprintf(stderr,"Ended parsing\nWill close listing file\n"); 
#else
  
	/* close listing file */
	if (xlist && list_level)
		fclose(lptr);
#endif
  #ifdef TEST
    fprintf(stderr,"closed list file\nwill close input file\n");
    #endif
  
	/* close input file */
	fclose(iptr);

  #ifdef TEST
    fprintf(stderr,"Will dump table\n");
    #endif
  
	/* dump the bank table */
	if (dump_seg)
		show_seg_usage();

	/* ok */
	return(0);
}

/* ----
 * show_seg_usage()
 * ----
 */

void
show_seg_usage(void)
{
	int i, j;
	int addr, start, nb;

	/* zp and bss usage */
	printf("Segment usage:\n");
	printf("\n");
	printf("      ZP    $%04X-$%04X\n",   RAM_BASE, RAM_BASE + (max_zp - 1));
	printf("     BSS    $%04X-$%04X\n\n", RAM_BASE, RAM_BASE + (max_bss - 1));

	/* bank usage */
	rom_used = 0;
	rom_free = 0;

	/* scan banks */
	for (i = 0; i <= max_bank; i++) {
		start = 0;
		addr = 0;
		nb = 0;

		/* count used and free bytes */
		for (j = 0; j < 8192; j++)
			if (map[i][j] != 0xFF)
				nb++;

		/* display bank infos */
		if (nb)			
			printf("bank% 4i    %20s    used:% 5i    free:% 5i\n",
					i, bank_name[i], nb, 8192 - nb);
		else {
			printf("bank% 4i    %20s    empty\n", i, bank_name[i]);
			continue;
		}

		/* update used/free counters */
		rom_used += nb;
		rom_free += 8192 - nb;

		/* scan */
		if (dump_seg == 1)
			continue;

		for (;;) {
			/* search section start */
			for (; addr < 8192; addr++)
				if (map[i][addr] != 0xFF)
					break;

			/* check for end of bank */
			if (addr > 8191)
				break;

			/* get section type */
			section = map[i][addr] & 0x0F;
			page = (map[i][addr] & 0xE0) << 8;
			start = addr;

			/* search section end */
			for (; addr < 8192; addr++)
				if ((map[i][addr] & 0x0F) != section)
					break;

			/* display section infos */
			printf("    %s    $%04X-$%04X\n",
					section_name[section],	/* section name */
				    start + page,			/* starting address */
					addr  + page - 1);		/* end address */
		}
	}

	/* total */
	rom_used = (rom_used + 1023) >> 10;
	rom_free = (rom_free) >> 10;
	printf("\t\t\t\t\t ------        ------\n");
	printf("\t\t\t\t\t % 5iK        % 5iK\n", rom_used, rom_free);
}

