#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "externs.h"
#include "protos.h"
#include "nes_inst.h"

unsigned char tile_plane[128];	/* tile plane buffer */
int ines_prg;		/* number of prg banks */
int ines_chr;		/* number of character banks */
int ines_mapper[2];	/* rom mapper type */
struct INES {		/* INES rom header */
	char	id[4];
	char	prg;
	char	chr;
	char	mapper[2];
	char	unused[8];
} header;

/* ----
 * init_machine()
 * ----
 * specific machine initialization
 */

void
init_machine(void)
{
	addinst(nes_pseudo);
}

/* ----
 * write_header()
 * ----
 * generate and write rom header
 */

void
write_header(FILE *f, int banks)
{
	/* setup INES header */
	memset(&header, 0, sizeof(header));
	header.id[0] = 'N';
	header.id[1] = 'E';
	header.id[2] = 'S';
	header.id[3] = 26;
	header.prg = ines_prg;
	header.chr = ines_chr;
	header.mapper[0] = ines_mapper[0];
	header.mapper[1] = ines_mapper[1];

	/* write */
	fwrite(&header, sizeof(header), 1, f);
}

/* ----
 * pack_8x8_tile()
 * ----
 * encode a 8x8 tile for the NES
 */

int
pack_8x8_tile(void *data, int format, int offset)
{
	int i, j;
	int cnt;
	unsigned int   pixel;
	unsigned char *ptr;
	unsigned int  *packed;

	/* pack the tile only in the last pass */
	if (pass == FIRST_PASS)
		return (16);

	/* clear planes */
	for (i = 0; i < 16; i++)
		tile_plane[i] = 0;

	/* encode the tile */
	switch (format) {
	case CHUNKY_TILE:
		/* 8-bit chunky format */
		cnt = 0;
		ptr = data;

		for (i = 0; i < 8; i++) {
			for (j = 0; j < 8; j++) {
				pixel = ptr[j ^ 0x07];
				tile_plane[cnt]    |= (pixel & 0x01) ? (1 << j) : 0;
				tile_plane[cnt+8]  |= (pixel & 0x02) ? (1 << j) : 0;
			}				
			ptr += offset;
			cnt += 1;
		}
		break;

	case PACKED_TILE:
		/* 4-bit packed format */
		cnt = 0;
		packed = data;
	
		for (i = 0; i < 8; i++) {
			pixel = packed[i];
	
			for (j = 0; j < 8; j++) {
				/* check for errors */
				if (pixel & 0x0C) {
					error("Incorrect pixel color index!");
					return (0);
				}

				/* convert the tile */
				tile_plane[cnt]    |= (pixel & 0x01) ? (1 << j) : 0;
				tile_plane[cnt+8]  |= (pixel & 0x02) ? (1 << j) : 0;
				pixel >>= 4;
			}
			cnt += 1;
		}
		break;
	}

	/* ok */
	return (16);
}

