#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "externs.h"
#include "protos.h"

extern int ines_prg;		/* number of prg banks */
extern int ines_chr;		/* number of character banks */
extern int ines_mapper[2];	/* rom mapper type */

/* ----
 * do_defchr()
 * ----
 * .defchr pseudo
 */

void
do_defchr(int *ip)
{
	unsigned int data[8];
	int i;

	/* define label */
	labldef(loccnt, 1);

	/* output infos */
	data_loccnt = loccnt;
	data_format = 3;
	data_level  = 3;

	/* get tile data */
	for (i = 0; i < 8; i++) {
		/* get value */
		if (!evaluate(ip, (i < 7) ? ',' : ';'))
			return;

		/* store value */
		data[i] = value;

		/* update location counter */
		loccnt += 2;
	}

	/* store tile on last pass */
	if (pass == LAST_PASS) {
		/* encode tile */
		if (!pack_8x8_tile(data, PACKED_TILE, 0))
			return;

		/* store planes */
		for (i = 0; i < 16; i++)
			putbyte(data_loccnt+i, tile_plane[i]);
	}

	/* output line */
	if (pass == LAST_PASS)
		println();
}

/* ----
 * do_inesprg()
 * ----
 * .inesprg pseudo
 */

void do_inesprg(int *ip)
{
	if (!evaluate(ip, ';'))
		return;

	if ((value < 0) || (value > 64)) 
	{
		error("Prg bank value out of range!");
	
		return;
	}
	
	ines_prg = value;

	if (pass == LAST_PASS) 
	{
		println();
	}
}

/* ----
 * do_ineschr()
 * ----
 * .ineschr pseudo
 */

void do_ineschr(int *ip)
{
	if (!evaluate(ip, ';'))
		return;

	if ((value < 0) || (value > 64)) 
	{
		error("Prg bank value out of range!");
	
		return;
	}
	
	ines_chr = value;

	if (pass == LAST_PASS) 
	{
		println();
	}
}

/* ----
 * do_inesmap()
 * ----
 * .inesmap pseudo
 */

void
do_inesmap(int *ip)
{
	if (!evaluate(ip, ';'))
		return;

	if ((value < 0) || (value > 255)) 
	{
		error("Mapper value out of range!");
	
		return;
	}
	
	ines_mapper[0] &= 0x0F;
	ines_mapper[0] |= (value & 0x0F) << 4;
	ines_mapper[1]  = (value & 0xF0);

	if (pass == LAST_PASS) 
	{
		println();
	}
}

/* ----
 * do_inesmir()
 * ----
 * .ines.mirror pseudo
 */

void
do_inesmir(int *ip)
{
	if (!evaluate(ip, ';'))
		return;

	if ((value < 0) || (value > 15)) 
	{
		error("Mirror value out of range!");
	
		return;
	}
	
	ines_mapper[0] &= 0xF0;
	ines_mapper[0] |= (value  & 0x0F);

	if (pass == LAST_PASS) 
	{
		println();
	}
}

