#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "externs.h"
#include "protos.h"

unsigned short buffer[8192];	/* buffer for .incbat and .incpal */

/* ----
 * do_vram()
 * ----
 * .vram pseudo
 */

void
do_vram(int *ip)
{
	/* define label */
	labldef(loccnt, 1);

	/* check if there's a label */
	if (lastlabl == NULL) {
		error("No label!");
		return;
	}

	/* get the VRAM address */
	if (!evaluate(ip, ';'))
		return;
	if (value >= 0x7F00) {
		error("Incorrect VRAM address!");
		return;
	}
	lastlabl->vram = value;

	/* output line */
	if (pass == LAST_PASS) {
		loadlc(value, 1);
		println();
	}
}

/* ----
 * do_pal()
 * ----
 * .pal pseudo
 */

void
do_pal(int *ip)
{
	/* define label */
	labldef(loccnt, 1);

	/* check if there's a label */
	if (lastlabl == NULL) {
		error("No label!");
		return;
	}

	/* get the palette index */
	if (!evaluate(ip, ';'))
		return;
	if (value > 15) {
		error("Incorrect palette index!");
		return;
	}
	lastlabl->pal = value;

	/* output line */
	if (pass == LAST_PASS) {
		loadlc(value, 1);
		println();
	}
}

/* ----
 * do_defpal()
 * ----
 * .defpal pseudo
 */

void
do_defpal(int *ip)
{
	unsigned int color;
	unsigned int r, g, b;
	char c;

	/* define label */
	labldef(loccnt, 1);

	/* output infos */
	data_loccnt = loccnt;
	data_format = 2;
	data_level  = 3;

	/* get data */
	for (;;) {
		/* get color */
		if (!evaluate(ip, 0))
			return;

		/* convert color */
		r = (value >> 8) & 0x7;
		g = (value >> 4) & 0x7;
		b = (value) & 0x7;
		color = (g << 6) + (r << 3) + (b);

		/* store data on last pass */
		if (pass == LAST_PASS)
			putword(loccnt, color);

		/* update location counter */
		loccnt += 2;

		/* check if there's another color */
		c = prlnbuf[(*ip)++];

		if (c != ',')
			break;
	}

	/* output line */
	if (pass == LAST_PASS)
		println();

	/* check errors */
	if (c != ';' && c != '\0')
		error("Syntax error!");
}

/* ----
 * do_defchr()
 * ----
 * .defchr pseudo
 */

void
do_defchr(int *ip)
{
	unsigned int data[8];
	int i;

	/* check if there's a label */
	if (lablptr == NULL) {
		error("No label to attach the char data to!");
		return;
	}

	/* define label */
	labldef(loccnt, 1);

	/* output infos */
	data_loccnt = loccnt;
	data_format = 3;
	data_level  = 3;

	/* get the VRAM address */
	if (!evaluate(ip, ','))
		return;
	if (value >= 0x7F00) {
		error("Incorrect VRAM address!");
		return;
	}
	lablptr->vram = value;

	/* get the default palette */
	if (!evaluate(ip, ','))
		return;
	if (value > 0x0F) {
		error("Incorrect palette index!");
		return;
	}
	lablptr->pal = value;

	/* get tile data */
	for (i = 0; i < 8; i++) {
		/* get value */
		if (!evaluate(ip, (i < 7) ? ',' : ';'))
			return;

		/* store value */
		data[i] = value;

		/* update location counter */
		loccnt += 4;
	}

	/* store tile on last pass */
	if (pass == LAST_PASS) {
		/* encode tile */
		if (!pack_8x8_tile(data, PACKED_TILE, 0))
			return;

		/* store planes */
		for (i = 0; i < 32; i++)
			putbyte(data_loccnt+i, tile_plane[i]);
	}

	/* output line */
	if (pass == LAST_PASS)
		println();
}

/* ----
 * do_defspr()
 * ----
 * .defspr pseudo
 */

void
do_defspr(int *ip)
{
	unsigned int data[32];
	int i;

	/* check if there's a label */
	if (lablptr == NULL) {
		error("No label to attach the sprite data to!");
		return;
	}

	/* define label */
	labldef(loccnt, 1);

	/* output infos */
	data_loccnt = loccnt;
	data_format = 3;
	data_level  = 3;

	/* get the VRAM address */
	if (!evaluate(ip, ','))
		return;
	if (value >= 0x7F00) {
		error("Incorrect VRAM address!");
		return;
	}
	lablptr->vram = value;

	/* get the default palette */
	if (!evaluate(ip, ','))
		return;
	if (value > 0x0F) {
		error("Incorrect palette index!");
		return;
	}
	lablptr->pal = value;

	/* get sprite data */
	for (i = 0; i < 32; i++) {
		/* get value */
		if (!evaluate(ip, (i < 31) ? ',' : ';'))
			return;

		/* store value */
		data[i] = value;

		/* update location counter */
		loccnt += 4;
	}

	/* store data on last pass */
	if (pass == LAST_PASS) {
		/* encode tile */
		if (!pack_16x16_tile(data, PACKED_TILE, 0))
			return;

		/* store planes */
		for (i = 0; i < 128; i++)
			putbyte(data_loccnt+i, tile_plane[i]);
	}

	/* output line */
	if (pass == LAST_PASS)
		println();
}

/* ----
 * do_incbat()
 * ----
 * build a BAT
 */

void
do_incbat(int n, unsigned int *arg)
{
	unsigned char *ptr, ref;
	unsigned int   i, j, k, l, x, y, w, h;
	unsigned int   base, index, flag = 0;

	/* get args */
	base = arg[0];
	x = 0;
	y = 0;

	if (n == 1) {		/* 1 arg */
		w = (pcx_w / 8);
		h = (pcx_h / 8);
	}
	else if (n == 3) {	/* 3 args */
		w = arg[1];
		h = arg[2];
	}
	else {				/* 5 args */
		x = arg[1];
		y = arg[2];
		w = arg[3];
		h = arg[4];
	}

	/* check coordinates validity */
	if (((x + w * 8) > pcx_w) || ((y + h * 8) > pcx_h)) {
		error("Coordinates out of range!");
		return;
	}

	/* build the BAT */
	if (pass == LAST_PASS) {
		index = 0;
		base  = (base >> 4);

		for (i = 0; i < h; i++) {
			for (j = 0; j < w; j++) {
				ptr = pcx_buf + (x + (j * 8)) + ((y + (i * 8)) * pcx_w);
				ref = ptr[0] & 0xF0;

				/* check colors */
				for (k = 0; k < 8; k++) {
					for (l = 0; l < 8; l++) {
						if ((ptr[l] & 0xF0) != ref)
							flag = 1;
					}				
					ptr += pcx_w;
				}
				buffer[index] = (base & 0xFFF) | ((ref & 0xF0) << 8);
				index++;
				base++;
			}
		}
	}
	putbuffer(buffer, 2 * w * h);

	if (flag)
		error("Invalid color index found!");
}

/* ----
 * do_incpal()
 * ----
 * extract the palette of a PCX file
 */

void
do_incpal(int n, unsigned int *arg)
{
	int i, start, nb;
	int r, g, b;

	/* get args */
	start = 0;
	nb = pcx_nb_colors;

	if (n) {
		start = arg[0] << 4;

		if (n == 1)
			nb = 16;
		else {
			nb = arg[1] << 4;
		}
	}

	/* check args */
	if (((start + nb) > 256) || (nb == 0)) {
		error("Palette index out of range!");
		return;
	}

	/* make the data */
	if (pass == LAST_PASS) {
		for (i = 0; i < nb; i++) {
			r = pcx_pal[start + i][0];
			g = pcx_pal[start + i][1];
			b = pcx_pal[start + i][2];
			buffer[i] = ((r & 0xE0) >> 2) | ((g & 0xE0) << 1) | ((b & 0xE0) >> 5);
		}
	}

	/* copy */
	putbuffer(buffer, nb << 1);
}


/* ----
 * do_incspr()
 * ----
 * PCX to sprites
 */

void
do_incspr(int n, unsigned int *arg)
{
	unsigned char *data;
	unsigned int i, j, x, y, w, h;
	unsigned int size;

	/* get args */
	if (!check_args(arg, n, &x, &y, &w, &h, 16))
		return;

	/* clear the tile plane buffer */
	memset(tile_plane, 0, 128);

	/* make the data */
	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++) {
			/* tile address in the PCX */
			data = pcx_buf + (x + (j << 4)) + ((y + (i << 4)) * pcx_w);

			/* encode the tile */
			size = pack_16x16_tile(data, CHUNKY_TILE, pcx_w);

			if (!size)
				return;

			/* store tile */
			putbuffer(tile_plane, size);
		}
	}

}

