#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "defs.h"
#include "externs.h"
#include "protos.h"

/* ----
 * symhash()
 * ----
 * calculate the hash value of a symbol
 */

int
symhash(void)
{
	int i;
	char c;
	int hash = 0;

	/* hash value */
	for (i = 1; i <= symbol[0]; i++) {
		c = symbol[i];
		hash += c;
		hash  = (hash << 3) + (hash >> 5) + c;
	}

	/* ok */
	return (hash & 0xFF);
}

/* ----
 * colsym()
 * ----
 * collect a symbol from prlnbuf into symbol[],
 * leaves prlnbuf pointer at first invalid symbol character,
 * returns 0 if no symbol collected
 */

int
colsym(int *ip)
{
	int	i;
	char c;

	/* get the symbol */
	i = 0;
	for (;;) {
		c = prlnbuf[*ip];
		if (isdigit(c) && (i == 0))
			break;
		if ((!isalnum(c)) && (c != '_') && (c != '.'))
			break;
		if (i < (SBOLSZ - 1))
			symbol[++i] = c;
		(*ip)++;
	}

	symbol[0] = i;
	symbol[i+1] = '\0';

	/* check if it's a reserved symbol */
	if (i == 1) {
		switch (toupper(symbol[1])) {
		case 'A':
		case 'X':
		case 'Y':
			i = -1;
			break;
		}
	}
	if (check_keyword())
		i = -1;

	/* check errors */
	if (i == -1) {
		error("Reserved symbol!");
		symbol[0] = 0;
		symbol[1] = '\0';
		i = 0;
	}	

	/* ok */
	return(i);
}

/* ----
 * stlook()
 * ----
 * symbol table lookup
 * if found, return pointer to symbol
 * else, install symbol as undefined and return pointer
 */

struct t_symbol *stlook(int flag)
{
	struct t_symbol *ptr;
	unsigned int hash = 0;
	char c;
	int	i;

	if (symbol[1] == '.'){		/* local symbol */
		if (glablptr) {
			ptr = glablptr->local;
			while (ptr) {
				if (!strcmp(symbol, ptr->name))
					break;			
				ptr = ptr->next;
			}
			if (ptr == NULL)
				if (flag)
					ptr = stinstal(hash, 1);
		}
		else {
			error("Local symbol not allowed here!");
			return (NULL);
		}
	} else {					/* global symbol */
		for (i = 0; i < symbol[0]; i++) {
			c = symbol[i + 1];
			hash += c;
			hash  = (hash << 3) + (hash >> 5) + c;
		}
		hash &= 0xFF;
		ptr = hash_tbl[hash];
		while (ptr) {
			if (!strcmp(symbol, ptr->name))
				break;			
			ptr = ptr->next;
		}
		if (ptr == NULL)
			if (flag)
				ptr = stinstal(hash, 0);
	}

	/* incremente symbol reference counter */
	if (ptr)
		ptr->refcnt++;

	/* ok */
	return(ptr);
}

/* ----
 * stinstall()
 * ----
 * install symbol into symbol hash table
 */

struct t_symbol *stinstal(int hash, int type)
{
	struct t_symbol *ptr;

	if ((ptr = (void *)malloc(sizeof(struct t_symbol))) == NULL) {
		fatal_error("Out of memory!");
		return (NULL);
	}
	ptr->type  = if_expr ? IFUNDEF : UNDEF;
	ptr->value = 0;
	ptr->local = NULL;
	ptr->bank  = -1;
	ptr->page  = -1;
	ptr->vram  = -1;
	ptr->pal   = -1;
	ptr->refcnt = 0;
	strcpy(ptr->name, symbol);

	if (type) {
		ptr->next = glablptr->local;
		glablptr->local = ptr;
	}
	else {
		ptr->next = hash_tbl[hash];
		hash_tbl[hash] = ptr;
	}
	return(ptr);
}

/* ----
 * labdef()
 * ----
 * assign <lval> to label pointed to by lablptr,
 * checking for valid definition, etc.
 */

int
labldef(int lval, int flag)
{
	char c;

	if (lablptr) {
		if (flag)
			lval = (lval & 0x1FFF) | (page << 13);

		/* first pass */
		if (pass == FIRST_PASS) {
			switch (lablptr->type) {
			case UNDEF:
				lablptr->type = DEFABS;
				lablptr->value = lval;
				break;

			case IFUNDEF:
				error("Can not define this label, declared as undefined in an IF expression!");
				return (-1);

			case MACRO:
				error("Symbol already used by a macro!");
				return (-1);

			case FUNC:
				error("Symbol already used by a function!");
				return (-1);

			default:
				lablptr->type = MDEF;
				lablptr->value = 0;
				error("Label multiply defined!");
				return (-1);
			}
		}

		/* second pass */
		else {
			if (lablptr->value != lval) {
				fatal_error("Internal error[1]!");
				return (-1);
			}
		}

		if (flag) {
			lablptr->bank = bank;
			lablptr->page = page;
			c = lablptr->name[1];
			if (c == '.')
				lastlabl = NULL;
			else {
				glablptr = lablptr;
				lastlabl = lablptr;
			}
		}
	}
	return(0);
}

