/*
 * Debug.c
 *
 * Creation : 30 Mar '00
 * For use in HuC
 *
 * Provides some functions to print informations on screen
 * Uses lots of stuff from the sample1.asm demo source from the magic kit
 */

/**[txh]********************************************************************

  Description: Initialize the debug font and such ...

  Return: nothing
  Example: init_debug_mode()   ^_^

***************************************************************************/


clr_debug_mode()
{
  #asm
	setvwaddr  $0		; set the VRAM address to $0000
        lda   #32               ; size 32 lines tall
.l1:    ldx   #64               ; size 64 chars wide
        pha
.l2:    cly

	; Fill each BAT map position with blank character

	stw   #CHAR(offchar),video_data

        dex                     ; next block
        bne   .l2
        pla
        dec   A                 ; next line
        bne   .l1
   #endasm
}

init_debug_mode()
{
#asm
        map    offchar		; map in the memory bank
        ; load blank character into VRAM:
        vload  offchar,#16

        set_bgpal  #0,cgpal,#1	; fill palette #0
; blank the background
#endasm

	clr_debug_mode();
#asm
	; load the characters for 0-9,A-F
	; into $1000 - $1100 in VRAM
	; (each character uses 16 words)

        ; 17 because I've temporary added the black character

        vload  zero,#16*17


#endasm

        return;

}

/**[txh]********************************************************************

  Description: Put a character on screen

  Return: nothing
  Example: putdigit(0xa, 0x80) will draw a 'A' on the first line, at the first
           place.
           add 0x40 to go one line down and 1 for 1 character right

***************************************************************************/

putdigit(character, offset)
int character;
int offset;
{
  #asm
    vreg  #$00

    __ldwp __stack
    stx   video_data
    sta   video_data+1

     __addmi 2, __stack

     vreg  #$02

     __ldwp __stack

     stx   video_data
     lda   #$01
     sta   video_data+1

     __submi 2, __stack

  #endasm

}

/**[txh]********************************************************************

  Description: Display a decimal number the position of the RIGHT character
      is given in offset

  Return: nothing
  Example: putnumber(15, 0x81) will display the string 15 beginning at the
    very first character on screen (@ 0x80, don't forget)

***************************************************************************/

putnumber(number, offset)
int number;
int offset;
{
 int w;
 do
 {
   /*
    We must debug the % operator which only works on Hu-Go!
    */

   w = (number / 10) * 10;
   putdigit(number - w, offset--);

   /*
     putdigit(number%10, offset--);
   */

   number = number / 10;
 }
 while (number);

 return;
}


     #asm

        .bank  MAIN_BANK+1
        .org   $4000

;
; Blank char
;
offchar:  .defchr $0FF0,0,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000

;
; numbers from 0-9,A-F:
;
zero:     .defchr $1000,0,\
	  $00111110,\
	  $01000011,\
	  $01000101,\
	  $01001001,\
	  $01010001,\
	  $01100001,\
	  $00111110,\
	  $00000000

one:      .defchr $1010,0,\
	  $00001000,\
	  $00011000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00011100,\
	  $00000000

two:      .defchr $1020,0,\
	  $00111110,\
	  $01000001,\
	  $00000001,\
	  $00111110,\
	  $01000000,\
	  $01000000,\
	  $01111111,\
	  $00000000

three:    .defchr $1030,0,\
	  $00111110,\
	  $01000001,\
	  $00000001,\
	  $00011110,\
	  $00000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

four:     .defchr $1040,0,\
	  $00000010,\
	  $00100010,\
	  $00100010,\
	  $00111111,\
	  $00000010,\
	  $00000010,\
	  $00000010,\
	  $00000000

five:     .defchr $1050,0,\
	  $01111111,\
	  $01000000,\
	  $01000000,\
	  $00111110,\
	  $00000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

six:      .defchr $1060,0,\
	  $00111110,\
	  $01000000,\
	  $01000000,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

seven:    .defchr $1070,0,\
	  $00111111,\
	  $00000001,\
	  $00000001,\
	  $00000010,\
	  $00000100,\
	  $00001000,\
	  $00001000,\
	  $00000000

eight:    .defchr $1080,0,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

nine:     .defchr $1090,0,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $00111111,\
	  $00000001,\
	  $00000001,\
	  $00111110,\
	  $00000000

ten:      .defchr $10A0,0,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $01111111,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $00000000

eleven:   .defchr $10B0,0,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $01111110,\
	  $00000000

twelve:   .defchr $10C0,0,\
	  $00111110,\
	  $01000001,\
	  $01000000,\
	  $01000000,\
	  $01000000,\
	  $01000001,\
	  $00111110,\
	  $00000000

thirteen: .defchr $10D0,0,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01111110,\
	  $00000000

fourteen: .defchr $10E0,0,\
	  $01111111,\
	  $01000000,\
	  $01000000,\
	  $01111110,\
	  $01000000,\
	  $01000000,\
	  $01111111,\
	  $00000000

fifteen:  .defchr $10F0,0,\
	  $01111111,\
	  $01000000,\
	  $01000000,\
	  $01111110,\
	  $01000000,\
	  $01000000,\
	  $01000000,\
	  $00000000

black:  .defchr $1100,0,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000


;
; Simple palette entry
;
; entry #0 = black, #1-#15 are all white
;
cgpal:  .defpal $000,$777,$777,$777,\
		$777,$777,$777,$777,\
		$777,$777,$777,$777,\
		$777,$777,$777,$777

;
; characters to place in BAT
;
; This simply prints all characters from 0-9, and A-F in a line
; $0100 = palette #0, pointing at $1000 for character (ie. '0')
;
chars:  dw   $0100,$0101,$0102,$0103
        dw   $0104,$0105,$0106,$0107
        dw   $0108,$0109,$010A,$010B
        dw   $010C,$010D,$010E,$010F

        .bank  MAIN_BANK

#endasm
