/*
 * Debug.c
 *
 * Creation : 30 Mar '00
 * For use in HuC
 *
 * Provides some functions to print informations on screen
 * Uses lots of stuff from the sample1.asm demo source from the magic kit
 */

#define CHAR_BLK 36

/**[txh]********************************************************************

  Description: Initialize the debug font and such ...

  Return: nothing
  Example: init_debug_mode()   ^_^

***************************************************************************/


clr_debug_mode()
{
  #asm
	setvwaddr  $0		; set the VRAM address to $0000
        lda   #32               ; size 32 lines tall
.l1:    ldx   #64               ; size 64 chars wide
        pha
.l2:    cly

	; Fill each BAT map position with blank character

	stw   #CHAR(offchar),video_data

        dex                     ; next block
        bne   .l2
        pla
        dec   A                 ; next line
        bne   .l1
   #endasm
}

init_debug_mode()
{
#asm
        map    offchar		; map in the memory bank
        ; load blank character into VRAM:
        vload  offchar,#16

        set_bgpal  #0,cgpal,#1	; fill palette #0
; blank the background
#endasm

	clr_debug_mode();
#asm
	; load the characters for 0-9,A-F
	; into $1000 - $1100 in VRAM
	; (each character uses 16 words)

        ; 37 because I've temporary added the black character

        vload  zero,#16*37


#endasm

        return;

}

/**[txh]********************************************************************

  Description: Put a character on screen

  Return: nothing
  Example: putdigit(0xa, 0x80) will draw a 'A' on the first line, at the first
           place.
           add 0x40 to go one line down and 1 for 1 character right

***************************************************************************/

putdigit(character, offset)
int character;
int offset;
{
  #asm
    vreg  #$00

    __ldwp __stack
    stx   video_data
    sta   video_data+1

     __addmi 2, __stack

     vreg  #$02

     __ldwp __stack
     __addwi 256

     stx   video_data
     ; lda   #$01
     sta   video_data+1

     __submi 2, __stack

  #endasm

}

/**[txh]********************************************************************

  Description: Display a decimal number the position of the RIGHT character
      is given in offset

  Return: nothing
  Example: putnumber(15, 0x81) will display the string 15 beginning at the
    very first character on screen (@ 0x80, don't forget)

***************************************************************************/

putnumber(number, offset)
int number;
int offset;
{
 int w;
 do
 {
   /*
    We must debug the % operator which only works on Hu-Go!
    */

   w = (number / 10) * 10;
   putdigit(number - w, offset--);

   /*
     putdigit(number%10, offset--);
   */

   number = number / 10;
 }
 while (number);

 return;
}

/**[txh]********************************************************************

  Description: Display an ascii string on screen

  Return: nothing
  Example: putstring("hello world");
           only digits, lower case letters and spaces are handled

***************************************************************************/

putstring(string,offset)
char* string;
int offset;
{
  char actual_char;

  while (*string)
    {
      actual_char = *string++;

      if (actual_char == ' ')
         actual_char = CHAR_BLK;
      else
      if (actual_char>='a')
         actual_char -= 87; /* 'a' = 97,
                                and 'a' is represented by character 10 */

      putdigit(actual_char, offset++);
    }
  return;
}


     #asm

        .bank  MAIN_BANK+1
        .org   $4000

;
; Blank char
;
offchar:  .defchr $0FF0,0,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000

;
; numbers from 0-9,A-F:
;
zero:     .defchr $1000,0,\
	  $00111110,\
	  $01000011,\
	  $01000101,\
	  $01001001,\
	  $01010001,\
	  $01100001,\
	  $00111110,\
	  $00000000

one:      .defchr $1010,0,\
	  $00001000,\
	  $00011000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00011100,\
	  $00000000

two:      .defchr $1020,0,\
	  $00111110,\
	  $01000001,\
	  $00000001,\
	  $00111110,\
	  $01000000,\
	  $01000000,\
	  $01111111,\
	  $00000000

three:    .defchr $1030,0,\
	  $00111110,\
	  $01000001,\
	  $00000001,\
	  $00011110,\
	  $00000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

four:     .defchr $1040,0,\
	  $00000010,\
	  $00100010,\
	  $00100010,\
	  $00111111,\
	  $00000010,\
	  $00000010,\
	  $00000010,\
	  $00000000

five:     .defchr $1050,0,\
	  $01111111,\
	  $01000000,\
	  $01000000,\
	  $00111110,\
	  $00000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

six:      .defchr $1060,0,\
	  $00111110,\
	  $01000000,\
	  $01000000,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

seven:    .defchr $1070,0,\
	  $00111111,\
	  $00000001,\
	  $00000001,\
	  $00000010,\
	  $00000100,\
	  $00001000,\
	  $00001000,\
	  $00000000

eight:    .defchr $1080,0,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

nine:     .defchr $1090,0,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $00111111,\
	  $00000001,\
	  $00000001,\
	  $00111110,\
	  $00000000

ten:      .defchr $10A0,0,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $01111111,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $00000000

eleven:   .defchr $10B0,0,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $01111110,\
	  $00000000

twelve:   .defchr $10C0,0,\
	  $00111110,\
	  $01000001,\
	  $01000000,\
	  $01000000,\
	  $01000000,\
	  $01000001,\
	  $00111110,\
	  $00000000

thirteen: .defchr $10D0,0,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01111110,\
	  $00000000

fourteen: .defchr $10E0,0,\
	  $01111111,\
	  $01000000,\
	  $01000000,\
	  $01111110,\
	  $01000000,\
	  $01000000,\
	  $01111111,\
	  $00000000

fifteen:  .defchr $10F0,0,\
	  $01111111,\
	  $01000000,\
	  $01000000,\
	  $01111110,\
	  $01000000,\
	  $01000000,\
	  $01000000,\
	  $00000000

char_g:  .defchr $1100,0,\
	  $00111110,\
	  $01000001,\
	  $01000000,\
	  $01000111,\
	  $01000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

char_h:  .defchr $1110,0,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01111111,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $00000000

char_i:  .defchr $1120,0,\
	  $00011100,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00011100,\
	  $00000000

char_j:  .defchr $1130,0,\
	  $00011110,\
	  $00000100,\
	  $00000100,\
	  $00000100,\
	  $00000100,\
	  $01000100,\
	  $00111000,\
	  $00000000

char_k:  .defchr $1140,0,\
	  $01000010,\
	  $01000100,\
	  $01001000,\
	  $01110000,\
	  $01001000,\
	  $01000100,\
	  $01000010,\
	  $00000000

char_l:  .defchr $1150,0,\
	  $01000000,\
	  $01000000,\
	  $01000000,\
	  $01000000,\
	  $01000000,\
	  $01000000,\
	  $01111110,\
	  $00000000

char_m:  .defchr $1160,0,\
	  $01000001,\
	  $01100011,\
	  $01010101,\
	  $01001001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $00000000

char_n:  .defchr $1170,0,\
	  $01000001,\
	  $01100001,\
	  $01010001,\
	  $01001001,\
	  $01000101,\
	  $01000011,\
	  $01000001,\
	  $00000000

char_o:  .defchr $1180,0,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

char_p:  .defchr $1190,0,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $01111110,\
	  $01000000,\
	  $01000000,\
	  $01000000,\
	  $00000000

char_q:  .defchr $11A0,0,\
	  $00111110,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000101,\
	  $01000011,\
	  $00111110,\
	  $00000000

char_r:  .defchr $11B0,0,\
	  $01111110,\
	  $01000001,\
	  $01000001,\
	  $01111110,\
	  $01000100,\
	  $01000010,\
	  $01000001,\
	  $00000000


char_s:  .defchr $11C0,0,\
	  $00111111,\
	  $01000000,\
	  $01000000,\
	  $00111110,\
	  $00000001,\
	  $00000001,\
	  $01111110,\
	  $00000000

char_t:  .defchr $11D0,0,\
	  $01111111,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00000000

char_u:  .defchr $11E0,0,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $00111110,\
	  $00000000

char_v:  .defchr $11F0,0,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $00100010,\
	  $00100010,\
	  $00011100,\
	  $00000000

char_w  .defchr $1200,0,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01000001,\
	  $01001001,\
	  $01001001,\
	  $00110110,\
	  $00000000

char_x  .defchr $1210,0,\
	  $01000001,\
	  $00100010,\
	  $00010100,\
	  $00001000,\
	  $00010100,\
	  $00100010,\
	  $01000001,\
	  $00000000

char_y  .defchr $1220,0,\
	  $01000001,\
	  $00100010,\
	  $00010100,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00001000,\
	  $00000000

char_z  .defchr $1230,0,\
	  $01111111,\
	  $00000010,\
	  $00000100,\
	  $00001000,\
	  $00010000,\
	  $00100000,\
	  $01111111,\
	  $00000000

char_blk:  .defchr $1240,0,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000,\
	  $00000000


;
; Simple palette entry
;
; entry #0 = black, #1-#15 are all white
;
cgpal:  .defpal $000,$777,$777,$777,\
		$777,$777,$777,$777,\
		$777,$777,$777,$777,\
		$777,$777,$777,$777

        .bank  MAIN_BANK

#endasm
