/*	File gen.c: 2.1 (83/03/20,16:02:06) */
/*% cc -O -c %
 *
 */

#include <stdio.h>
#include "defs.h"
#include "data.h"

/*
 *	return next available internal label number
 *
 */
getlabel ()
{
	return (nxtlab++);
}

/*
 *	fetch a static memory cell into the primary register
 */
getmem (sym)
char *sym;
{
	if ((sym[IDENT] != POINTER) & (sym[TYPE] == CCHAR)) {
		out_ins(I_LDB,  T_SYMBOL, sym + NAME);
		out_ins(I_EXTW, NULL, NULL);
	} else {
		out_ins(I_LDW,  T_SYMBOL, sym + NAME);
	}
}

/*
 *	fetch the address of the specified symbol into the primary register
 *
 */
getloc (sym)
char *sym;
{
	int value;

	if (sym[STORAGE] == LSTATIC)
		out_ins(I_LDWI, T_LABEL, glint(sym));
	else {
		value = glint(sym) - stkp;
		out_ins(I_LDW, T_STACK, NULL);

		if (value) {
			out_ins(I_ADDWI, T_VALUE, value);
		}
	}
}

/*
 *	store the primary register into the specified static memory cell
 *
 */
putmem (sym)
char *sym;
{
	if ((sym[IDENT] != POINTER) & (sym[TYPE] == CCHAR))
		out_ins(I_STB, T_SYMBOL, sym + NAME);
	else
		out_ins(I_STW, T_SYMBOL, sym + NAME);
}

/*
 *	store the specified object type in the primary register
 *	at the address on the top of the stack
 *
 */
putstk (typeobj)
char typeobj;
{
	if (typeobj == CCHAR)
		out_ins(I_STBPS, NULL, NULL);
	else
		out_ins(I_STWPS, NULL, NULL);
	stkp = stkp + INTSIZE;
}

/*
 *	fetch the specified object type indirect through the primary
 *	register into the primary register
 *
 */
indirect (typeobj)
char typeobj;
{
	out_ins(I_STW, T_PTR, NULL);
	if (typeobj == CCHAR) {
		out_ins(I_LDBP, T_PTR, NULL);
		out_ins(I_EXTW, NULL, NULL);
	} else {
		out_ins(I_LDWP, T_PTR, NULL);
	}
}

/*
 *	print partial instruction to get an immediate value into
 *	the primary register
 *
 */
immed (type, data)
int type, data;
{
	out_ins(I_LDWI, type, data);
}

/*
 *	push the primary register onto the stack
 *
 */
gpush ()
{
	out_ins(I_PUSHW, NULL, NULL);
	stkp = stkp - INTSIZE;
}

/*
 *	pop the top of the stack into the secondary register
 *
 */
gpop ()
{
	out_ins(I_POPW, NULL, NULL);
	stkp = stkp + INTSIZE;
}

/*
 *	swap the primary register and the top of the stack
 *
 */
swapstk ()
{
	out_ins(I_SWAPW, NULL, NULL);
}

/*
 *	call the specified subroutine name
 *
 */
gcall (sname)
char *sname;
{
	out_ins(I_JSR, T_SYMBOL, sname);
}

/*
 *	return from subroutine
 *
 */
gret ()
{
	out_ins(I_RTS, NULL, NULL);
}

/*
 *	perform subroutine call to value on top of stack
 *
 */
callstk ()
{
	out_ins(I_CALLS, NULL, NULL);
}

/*
 *	jump to specified internal label number
 *
 */
jump (label)
int	label;
{
	out_ins(I_LBRA, T_LABEL, label);
}

/*
 *	test the primary register and jump if false to label
 *
 */
testjump (label, ft)
int	label, ft;
{
	out_ins(I_TSTW, NULL, NULL);
	if (ft)
		out_ins(I_LBNE, T_LABEL, label);
	else
		out_ins(I_LBEQ, T_LABEL, label);
}

/*
 *	modify the stack pointer to the new value indicated
 *
 */
modstk (newstkp)
int	newstkp;
{
	int	k;

//	k = galign(newstkp - stkp);
	k = newstkp - stkp;
	if (k)
		out_ins(I_ADDMI, T_STACK, k);
	return (newstkp);
}

/*
 *	multiply the primary register by INTSIZE
 */
gaslint ()
{
	out_ins(I_ASLW, NULL, NULL);
}

/*
 *	divide the primary register by INTSIZE
 */
gasrint()
{
	out_ins(I_ASRW, NULL, NULL);
}

/*
 *	Case jump instruction
 */
gjcase() {
	out_ins(I_JMP, T_SYMBOL, "__case");
}

/*
 *	add the primary and secondary registers
 *	if lval2 is int pointer and lval is int, scale lval
 */
gadd (lval, lval2) int *lval, *lval2;
{
	if (dbltest (lval2, lval)) {
		out_ins(I_ASLWS, NULL, NULL);
	}
	out_ins(I_ADDWS, NULL, NULL);
	stkp = stkp + INTSIZE;
}

/*
 *	subtract the primary register from the secondary
 *
 */
gsub ()
{
	out_ins(I_SUBWS, NULL, NULL);
	stkp = stkp + INTSIZE;
}

/*
 *	multiply the primary and secondary registers
 *	(result in primary)
 *
 */
gmult ()
{
	out_ins(I_JSR, T_LIB, "smul");
	stkp = stkp + INTSIZE;
}

/*
 *	divide the secondary register by the primary
 *	(quotient in primary, remainder in secondary)
 *
 */
gdiv ()
{
	out_ins(I_JSR, T_LIB, "sdiv");
	stkp = stkp + INTSIZE;
}

/*
 *	compute the remainder (mod) of the secondary register
 *	divided by the primary register
 *	(remainder in primary, quotient in secondary)
 *
 */
gmod ()
{
	out_ins(I_JSR, T_LIB, "smod");
	stkp = stkp + INTSIZE;
}

/*
 *	inclusive 'or' the primary and secondary registers
 *
 */
gor ()
{
	out_ins(I_ORWS, NULL, NULL);
	stkp = stkp + INTSIZE;
}

/*
 *	exclusive 'or' the primary and secondary registers
 *
 */
gxor ()
{
	out_ins(I_EORWS, NULL, NULL);
	stkp = stkp + INTSIZE;
}

/*
 *	'and' the primary and secondary registers
 *
 */
gand ()
{
	out_ins(I_ANDWS, NULL, NULL);
	stkp = stkp + INTSIZE;
}

/*
 *	arithmetic shift right the secondary register the number of
 *	times in the primary register
 *	(results in primary register)
 *
 */
gasr ()
{
	out_ins(I_JSR, T_LIB, "asr");
	stkp = stkp + INTSIZE;
}

/*
 *	arithmetic shift left the secondary register the number of
 *	times in the primary register
 *	(results in primary register)
 *
 */
gasl ()
{
	out_ins(I_JSR, T_LIB, "asl");
	stkp = stkp + INTSIZE;
}

/*
 *	two's complement of primary register
 *
 */
gneg ()
{
	out_ins(I_NEGW, NULL, NULL);
}

/*
 *	one's complement of primary register
 *
 */
gcom ()
{
	out_ins(I_COMW, NULL, NULL);
}

/*
 *	convert primary register into logical value
 *
 */
gbool ()
{
	out_ins(I_BOOLW, NULL, NULL);
}

/*
 *	logical complement of primary register
 *
 */
glneg ()
{
	out_ins(I_NOTW, NULL, NULL);
}

/*
 *	increment the primary register by 1 if char, INTSIZE if
 *      int
 */
ginc (lval) int lval[];
{
	if (lval[2] == CINT)
		out_ins(I_ADDWI, T_VALUE, 2);
	else
		out_ins(I_ADDWI, T_VALUE, 1);
}

/*
 *	decrement the primary register by one if char, INTSIZE if
 *	int
 */
gdec (lval) int lval[];
{
	if (lval[2] == CINT)
		out_ins(I_SUBWI, T_VALUE, 2);
	else
		out_ins(I_SUBWI, T_VALUE, 1);
}

/*
 *	following are the conditional operators.
 *	they compare the secondary register against the primary register
 *	and put a literl 1 in the primary if the condition is true,
 *	otherwise they clear the primary register
 *
 */

/*
 *	equal
 *
 */
geq ()
{
	out_ins(I_JSR, T_LIB, "eq");
	stkp = stkp + INTSIZE;
}

/*
 *	not equal
 *
 */
gne ()
{
	out_ins(I_JSR, T_LIB, "ne");
	stkp = stkp + INTSIZE;
}

/*
 *	less than (signed)
 *
 */
glt ()
{
	out_ins(I_JSR, T_LIB, "lt");
	stkp = stkp + INTSIZE;
}
/*
 *	less than or equal (signed)
 *
 */
gle ()
{
	out_ins(I_JSR, T_LIB, "le");
	stkp = stkp + INTSIZE;
}

/*
 *	greater than (signed)
 *
 */
ggt ()
{
	out_ins(I_JSR, T_LIB, "gt");
	stkp = stkp + INTSIZE;
}

/*
 *	greater than or equal (signed)
 *
 */
gge ()
{
	out_ins(I_JSR, T_LIB, "ge");
	stkp = stkp + INTSIZE;
}

/*
 *	less than (unsigned)
 *
 */
gult ()
{
	out_ins(I_JSR, T_LIB, "ult");
	stkp = stkp + INTSIZE;
}

/*
 *	less than or equal (unsigned)
 *
 */
gule ()
{
	out_ins(I_JSR, T_LIB, "ule");
	stkp = stkp + INTSIZE;
}

/*
 *	greater than (unsigned)
 *
 */
gugt ()
{
	out_ins(I_JSR, T_LIB, "ugt");
	stkp = stkp + INTSIZE;
}

/*
 *	greater than or equal (unsigned)
 *
 */
guge ()
{
	out_ins(I_JSR, T_LIB, "uge");
	stkp = stkp + INTSIZE;
}

