
; ----
; ResetPSG();
; ----
; Reset the sound generator.
; ----

_ResetPSG:
	stz   $0801		; main volume to zero
	stz   $0809		; disable the LFO
	
	lda   #5		; set volume to zero for each channel (0-5)
.clear:	sta   $0800
	stz   $0804
	dec   A
	bpl   .clear
	rts

; ----
; SetupDisplay(char mode, char disp_w, char disp_h, char v_w, char v_h);
; ----
; Set up a screen. If the values are incorrect, use a default
; screen size of 32x30 (note: all the sizes are in characters)
; ----
; mode:   0 = 256 pixels res. (5MHz), 1 = 320 pixels res. (7MHz)
; disp_w: horizontal size of the display, 10 - 34 if 5MHz clock
;                                         16 - 44 if 7MHz clock
; disp_h: vertical size, 10 - 30
; v_w:    virtual horizontal size, 0 - 3, 0 = 32
;                                         1 = 64
;                                         2 = 128
;                                         3 = 128
; v_h:    vertical size, 0 - 1, 0 = 32
;                               1 = 64
; ----

_SetupDisplay:
	__popw
	  stx	<_al
	__popw
	  stx	<_dl
	__popw
	  stx	<_dh
	__popw
	  stx	<_cl
	__popw
	  stx	<_ch

	  lda	<_dh		; check the height
	  cmp	#10
	  blo	.default
	  cmp	#31
	  bhs	.default
	  tst	#1,<_al
	  bne	._7MHz

;- 5MHz frequency (256 pixels)

._5MHz:	  lda	<_dl
	  cmp	#10
	  blo	.default
	  cmp	#35
	  bhs	.default

.set:	  vreg	#10		; HSR
	  st1	#2
	  lda	<_dl
	  tax
	  dex
	  lsr	A
	  neg	
	  add	#$12
	  sta	$0003
	  add	<_dl
	  neg
	  add	#$26
	  tay
	  vreg	#11		; HDR
	  stx	$0002
	  sty	$0003

	  clx
	  bra	.mode

;- default screen mode

.default: stz	<_al
	  stwz	<_cl
          ; Replaced stwi by stw thanks to new magic kit 2.0 features
          ; stwi  $1E20,<_dl
          stw $1E20,<_dl
	  bra	.set

;- 7MHz frequency (320 pixels)

._7MHz:	  lda	<_dl
	  cmp	#16
	  blo	.default
	  cmp	#45
	  bhs	.default

	  vreg	#10		; HSR
	  st1	#3
	  lda	<_dl
	  tax
	  dex
	  lsr	A
	  neg
	  add	#$18
	  sta	$0003
	  clc
	  adc	<_dl
	  neg
	  add	#$34
	  tay
	  vreg	#11		; HDR
	  stx	$0002
	  sty	$0003

	  ldx	#$1A

;- mode

.mode:	  vreg	#9
	  stx	$0002

;- vertical display

.vset:	  vreg	#12		; VPR
	  st1	#2
	  lda	<_dh
	  asl	A
	  asl	A
	  asl	A
	  tax
	  dex
	  lsr	A
	  neg
	  add	#135
	  sta	$0003
	  vreg	#13		; VDR
	  stx	$0002
	  stz	$0003
	  vreg	#14		; VCR
	  txa
	  inc	A
	  inc	A
	  lsr	A
	  adc	#125
	  sta	$0002
	  stz	$0003

;- clock frequency

	  lda	<_al
	  eor	#4
	  sta	$0400
	  rts


; SetPalette(char index, int color);
; ---------------------------------
; Set one entry of the palette to the specified color.
; ----
; index: entry index in the palette (0-511)
; color: color value,  GREEN: bit 6-8
;                      RED:   bit 3-5
;                      BLUE:  bit 0-2
; ----

_SetPalette:
	__popw
	__stw	$0402
	__popw
	__stw	$0404
	  rts


; LoadPalette(char block, int *data, char nb);
; -------------------------------------------
; Load # block(s) of 16 colors into the palette.
; ----
; block: block index in the palette (0-31)
; data:  pointer to a block of 16 colors
; nb:    number of block to upload (1-32)
; ----

LoadPalette:
	__popw
	  stx	<_al
	__popw
	__stw	<_si
	__popw
	  stx	<_cl
	  lda	<_al		; multiply the block index by 16
	  stz   <_ah		; and set the palette index
	  asl   A
	  asl   A
	  asl   A
	  asl   A
	  rol   <_ah
	  sta   $0402
	  lda   <_ah
	  sta   $0403

	  cly
.loop_a:
	  ldx	#16
.loop_b:
	  lda	[_si],Y		; set a color
	  iny
	  sta	$0404
	  lda	[_si],Y
	  sta	$0405
	  iny
	  bne	.next
	  inc	<_si+1
.next:    dex			; next color
	  bne	.loop_b

	  dec	<_cl		; next block
	  bne	.loop_a
	  rts


; vbl_on:
; ------
; enable the vbl interrupt of the VDC

vbl_on:
	smb3  <_creg
	vreg  #$5
	lda   <_creg
	sta   $0002
	rts


; vbl_off:
; -------
; disable the vbl interrupt of the VDC

vbl_off:
	rmb3  <_creg
	vreg  #$5
	lda   <_creg
	sta   $0002
	rts


; set_sprite:
; ----------
; change the attributes of a sprite
;
; IN:   A = sprite number (0-63) 
;       _SI = pointer to a SATB entry 
;       _CL = number of sprites to copy
;
; NOTE: this procedure assumes that the SATB is located
;       at address $7F00 in the VRAM

set_sprite:
	asl   A
	asl   A
	sta   <_di
	lda   #$7F
	sta   <_di+1
	lda   <_cl
	asl   A
	asl   A
	sta   <_cl
	stz   <_ch
	jmp   load_vram


; load_vram:
; ---------
; copy a block of memory to the VRAM
;
; IN:   _DI = VRAM location
;       _SI = memory location
;       _CX = number of words to copy
;
;load_vram:
;	jsr   set_write
;	cly
;.loop:	lda   [_si],Y
;	sta   $0002
;	iny
;	lda   [_si],Y
;	sta   $0003
;	iny
;	bne   .next
;	inc   <_si+1
;.next:  decw  <_cx
;	lda   <_cl
;	ora   <_ch
;	bne   .loop
;	rts
;
;
; set_read:
; ---------
; initialize the VRAM for reading
;
; IN:   _DI = VRAM location
;
;set_read:
;	vreg  #$01
;	lda   <_di 
;	sta   $0002
;	lda   <_di+1
;	sta   $0003
;	vreg  #$02
;	rts 
;
;
; set_write:
; ---------
; initialize the VRAM for writing
;
; IN:   _DI = VRAM location
;
;set_write:
;	vreg  #$00
;	lda   <_di 
;	sta   $0002
;	lda   <_di+1
;	sta   $0003
;	vreg  #$02
;	rts 

